import os
import glob
import numpy as np
import pandas as pd
import h5py
from intervalframe.write.write_h5 import write_h5_intervalframe


def convert_recarray(rec_array):
    """
    Convert the dtypes of np.recarray to h5py compatible dtypes

    Parameters
    ----------
        rec_array
            numpy.recarray (Array to convert dtypes for)

    Returns
    -------
        rec_array
            numpy.recarray (Array with h5py compatible dtypes)
    """

    dtypes = {o:rec_array.dtype[o].str for o in rec_array.dtype.fields}
    for key_dtype in dtypes:
        if dtypes[key_dtype] == "|O":
            lengths = []
            for i in range(len(rec_array[key_dtype])):
                if isinstance(rec_array[key_dtype][i], str):
                    lengths.append(len(rec_array[key_dtype][i]))
                elif pd.isnull(rec_array[key_dtype][i]):
                    continue
                elif isinstance(rec_array[key_dtype][i], int):
                    continue
                elif isinstance(rec_array[key_dtype][i], float):
                    continue
                else:
                    continue
            max_length = max(lengths) if len(lengths) > 0 else 0
            if max_length > 0:
                dtypes[key_dtype] = "<S{}".format(max_length)
            else:
                dtypes[key_dtype] = "<f8"

    rec_array = rec_array.astype(list(dtypes.items()))

    return rec_array


def write_h5_DataFrame(h5_group, df, axis=0):
    """
    Write pandas.DataFrame to h5py group

    Parameters
    ----------
        h5_group
            h5py.group
        df
            pandas.DataFrame
        axis
            int

    Returns
    -------
        None
    """
    
    # Record axis
    h5_group["axis"] = axis
    h5_group["shape"] = np.array(df.shape)

    # Handle MultiIndex
    if isinstance(df.columns, pd.MultiIndex):
        h5_group["columns_type"] = np.array([b"multi"])
    else:
        h5_group["columns_type"] = np.array([b"single"])

    # Iterate over columns
    if axis == 1:
        if df.index.dtype.kind == "i":
            index_dtypes = "i"
        else:
            index_dtypes = "<S{}".format(df.index.str.len().max())
        
        #if df.columns.dtype.kind == "i":
        #    columns_dtypes = "<S{}".format(df.columns.str.len().max())
        #else:
        #    columns_dtypes = "<S{}".format(df.columns.str.len().max())

        rec_array = df.T.to_records(index_dtypes=index_dtypes)

        rec_array = convert_recarray(rec_array)

        h5_group["values"] = rec_array
    
    # Iterate over index
    if axis == 0:
        if df.index.dtype.kind == "i":
            index_dtypes = "i"
        else:
            index_dtypes = "<S{}".format(df.index.str.len().max())
        if df.columns.dtype.kind == "i":
            columns_dtypes = "i"
        else:
            columns_dtypes = "<S{}".format(df.columns.str.len().max())

        rec_array = df.to_records(index_dtypes=index_dtypes)

        rec_array = convert_recarray(rec_array)

        h5_group["values"] = rec_array


def write_h5(cfdna_object, h5_filename):
    """
    """
    
    # Open file
    f = h5py.File(h5_filename, "w")

    # Write anno
    anno = f.create_group("anno")
    write_h5_DataFrame(anno, cfdna_object.anno, axis=1)

    # Write intervals
    intervals = f.create_group("intervals")
    for key in cfdna_object.intervals:
        key_intervals = intervals.create_group(key)
        write_h5_intervalframe(cfdna_object.intervals[key], key_intervals)

    # Write obs intervals
    obs_intervals = f.create_group("obs_intervals")
    for key in cfdna_object.obs_intervals:
        #key_obs_intervals = obs_intervals.create_group(key+"_obs_intervals")
        key_obs_intervals = obs_intervals.create_group(key)
        for obs in cfdna_object.obs_intervals[key]:
            #key_obs_obs_intervals = key_obs_intervals.create_group(key+"_obs_obs_intervals")
            key_obs_obs_intervals = key_obs_intervals.create_group(obs)
            write_h5_intervalframe(cfdna_object.obs_intervals[key][obs], key_obs_obs_intervals)

    # Write obs values
    obs_values = f.create_group("obs_values")
    for key in cfdna_object.obs_values:
        #key_obs_values = obs_values.create_group(key+"obs_values")
        key_obs_values = obs_values.create_group(key)
        write_h5_DataFrame(key_obs_values, cfdna_object.obs_values[key], axis=1)

    # Write chroms
    f["chroms"] = np.array(list(cfdna_object.chroms)).astype(bytes)

    # Write hmm
    hmm_states = f.create_group("hmm_states")
    for key in cfdna_object.hmm_states:
        #key_hmm_states = hmm_states.create_group(key+"_hmm_states")
        key_hmm_states = hmm_states.create_group(key)
        key_hmm_states["mus"] = cfdna_object.hmm_states[key]["mus"]
        key_hmm_states["lambdas"] = cfdna_object.hmm_states[key]["lambdas"]
        key_hmm_states["nu"] = np.array([cfdna_object.hmm_states[key]["nu"]])
        key_hmm_states["states"] = cfdna_object.hmm_states[key]["states"].values
        key_hmm_states["subclone"] = cfdna_object.hmm_states[key]["subclone"].values
        key_hmm_states["n"] = np.array([cfdna_object.hmm_states[key]["n"]])
        key_hmm_states["phi"] = np.array([cfdna_object.hmm_states[key]["phi"]])
        key_hmm_states["sp"] = np.array([cfdna_object.hmm_states[key]["sp"]])
        key_hmm_states["Frac_genome_subclonal"] = np.array([cfdna_object.hmm_states[key]["Frac_genome_subclonal"]])
        #cn_key_hmm_states = hmm_states.create_group("cn_"+key+"_hmm_states")
        cn_key_hmm_states = key_hmm_states.create_group("cn")
        write_h5_intervalframe(cfdna_object.hmm_states[key]["cn"], cn_key_hmm_states)

    # Write file names
    f["filenames"] = np.array(list(cfdna_object.filenames)).astype(bytes)

    # Write bin sizes
    if cfdna_object.cnv_binsize is not None:
        f["cnv_binsize"] = np.array([cfdna_object.cnv_binsize])
    if cfdna_object.hmm_binsize is not None:
        f["hmm_binsize"] = np.array([cfdna_object.hmm_binsize])
    if cfdna_object.profile_binsize is not None:
        f["profile_binsize"] = np.array([cfdna_object.profile_binsize])

    # Write chrom lengths
    chrom_lengths = f.create_group("chrom_lengths")
    chrom_lengths["chroms"] = np.array(list(cfdna_object.chrom_lengths.keys())).astype(bytes)
    chrom_lengths["lengths"] = np.array(list(cfdna_object.chrom_lengths.values()))

    # Close file
    f.close()


from .cnv.segmentation import *
from .nucleosome.nfr import *
from .nucleosome.wps import *
from .fragmentation.frag_pattern import *
from .summarize.summarize import *


__doc__="""\

Processing API
============
.. currentmodule:: cfDNA

Generic
-------
.. autosummary::
   :toctree: .

   segmentation
   nfr
   wps
   frag_pattern
   summarize

"""


import os
import numpy as np
import pandas as pd
import h5py
import argparse
from intervalframe import IntervalFrame
from intervalframe.write.write_h5 import write_h5_intervalframe
from intervalframe.read.read_h5 import read_h5_intervalframe
from ngsfragments import fragments, utilities
from ngsfragments.segment.cnv import call_cnvs, process_cnvs
from ngsfragments.segment import cnv_utilities
from ngsfragments.segment import correction


class cfDNA(object):
    """
    Wrapper of fragments for cell-free DNA

	:class:`~cfDNA.cfDNA` stores a fragments object

    """

    def __init__(self, frags=None, cfdna_h5=None, ref_genome="hg19", verbose=False):
        """
        Initialize cfDNA object
                 
        Parameters
        ----------
            frags : :class: `ngsfragments.ngsfragments` (default: None)
                DNA fragment intervals from BAM file
            ref_genome : str
                Reference genome
            verbose : bool
                 Print process
                 
        Returns
        -------
            None

        """

        if cfdna_h5 is None:
            # Initialize analysis properties
            self.ref_genome = ref_genome
            self.gender = {}

            self.filenames = set()
            self.chrom_lengths = None
            self.cnv_binsize = None
            self.hmm_binsize = None
            self.profile_binsize = None
            self.hmm_states = {}
            self.chroms = None

            self._anno = None
            self._intervals = {}
            self._obs_intervals = {}
            self._obs_values = {}
        else:
            pass

        # Initialize values from fragments
        if frags is not None:
            self.log_filename(frags)


    @property
    def anno(self):
        """
        pandas.DataFrame: Observation annotations
        """

        return self._anno

    @property
    def intervals(self):
        """
        dict of IntervalFrames: Shared interval values
        """

        return self._intervals


    @property
    def obs_intervals(self):
        """
        dict of IntervalFrames: Observation specific interval values
        """

        return self._obs_intervals


    @property
    def obs_values(self):
        """
        dict of pandas.DataFrame: Shared values
        """

        return self._obs_values


    def __len__(self):
        """
        Get length from self.frags
        """

        return len(self.filenames)


    def add_anno(self, key, file_name, value):
        """
        Add value to key in .anno

        Parameters
        ----------
            key : str
                Key for .anno columns
            file_name : str
                Observation file name
            value : str, float, or int
                Values to record

        Returns
        -------
            None

        """

        if self._anno is None:
            self._anno = pd.DataFrame(value, index=[file_name], columns=[key])
        self._anno.loc[file_name, key] = value


    def add_intervals(self, key, file_name, values):
        """
        Add value to key in .anno

        Parameters
        ----------
            key : str
                Key for .anno columns
            file_name : str or list-like
                Observation file name
            values : IntervalFrame
                Values to record

        Returns
        -------
            None

        """

        try:
            self._intervals[key]
            values = values.exact_match(self._intervals[key])
            self._intervals[key] = self._intervals[key].exact_match(values)
            self._intervals[key].df.loc[:,file_name] = values.loc[:,file_name].values

        except KeyError:
            self._intervals[key] = values

        
    def add_obs_intervals(self, key, file_name, values):
        """
        Add value to key in .anno

        Parameters
        ----------
            key : str
                Key for .anno columns
            file_name : str
                Observation file name
            values : IntervalFrame
                Values to record

        Returns
        -------
            None

        """

        try:
            self._obs_intervals[key]
        except KeyError:
            self._obs_intervals[key] = {}
        self._obs_intervals[key][file_name] = values


    def add_obs_values(self, key, file_name, values):
        """
        Add value to key in .anno

        Parameters
        ----------
            key : str
                Key for .anno columns
            file_name : str
                Observation file name
            values : np.array or pandas.DataFrame
                Values to record

        Returns
        -------
            None

        """

        # Check shape
        if values.shape[0] == len(self):
            pass
        elif len(values.shape) == 2:
            if values.shape[1] == len(self):
                values = values.T
        elif len(self) == 1:
            pass
        elif len(values.shape) == 1:
            pass
        else:
            raise IndexError("Shape of values cannot be matched.")

        try:
            self._obs_values[key]
            if isinstance(values, np.ndarray):
                self._obs_values[key].loc[file_name,:] = values
            elif isinstance(values, pd.DataFrame):
                self._obs_values[key] = pd.concat([self._obs_values[key], values]).drop_duplicates()
        except KeyError:
            if isinstance(values, np.ndarray):
                self._obs_values[key] = pd.DataFrame(values, columns=[file_name]).T
            elif isinstance(values, pd.DataFrame):
                self._obs_values[key] = values


    def log_filename(self, frags):
        """
        """

        # Determine name
        path = os.path.normpath(frags.sam_file)
        file_name = path.split(os.sep)[-1]

        # Find all file names
        if self._anno is None:
            total_filenames = np.array([file_name])
        else:
            total_filenames = np.append(self._anno.index.values, file_name)
            self._anno = self._anno.reindex(total_filenames)

        # Record
        self.filenames.add(file_name)
        self.add_obs_values("length_dist", file_name, frags.length_dist())
        self.add_anno("n_fragments", file_name, len(frags.frags))

        if self.chrom_lengths is None:
            self.chrom_lengths = frags.genome
        if self.chroms is None:
            self.chroms = frags.chroms

        



def wps(frags, protection=120, min_length=120, max_length=210, normalize=True, norm_method="mean"):
    """
    Calculate Window Protection Score (WPS)

    Parameters
    ----------
        chrom : str
            Name of chromosome
        protection : int
            Protection window
        min_length : int
            Minimum DNA fragment length
        max_length : int
            Maximum fragment length
        normalize : bool
            Whether to normalize
        norm_method : str
            Normalization method

    Returns
    -------
        wps : pandas.Series
            Window protection scores
    """
    
    # Caclulate WPS
    wps = frags.wps(protection=protection, min_length=min_length, max_length=max_length)
    if normalize:
        utilities.normalize_wps(wps, method=norm_method)

    return wps


def predict_nucleosomes(self, fragments, wps=None, protection=120, merge_distance=5,
				            min_length=120, max_length=210):
    """
    Call peaks from WPS

    Parameters
    ----------
        chrom
            str (chromosomes to call for)

    Returns
    -------
        peaks
            dict[chrom:AIList] (WPS peaks)
    """

    # Calculate wps
    if wps is None:
        peaks = {}
        for chrom in self.chroms:
            wps = fragments.wps(chrom=chrom, protection=protection, min_length=min_length, max_length=max_length)
            utilities.normalize_wps(wps)
            tmp_peaks = utilities.wps_peaks(fragments, wps, merge_distance=merge_distance, min_length=min_length, max_length=max_length)
            peaks[chrom] = tmp_peaks[chrom]
    else:
        peaks = utilities.wps_peaks(fragments, wps, merge_distance=merge_distance, min_length=min_length, max_length=max_length)

    return peaks


def peak_distances(self, fragments, peaks=None, bin_size=100000, max_distance=10000):
    """
    Determine distance between peaks

    Parameters
    ----------
        peaks
            dict[chrom:AIList]
        bin_size
            int
        max_distance
            int
        smooth
            bool

    Returns
    -------
        peak_bins
            dict[pandas.Series]
    """

    # Calculate peaks
    if peaks is None:
        peaks = self.predict_nucleosomes(fragments)

    # Create p_dist IntervalFrame
    starts = np.array([],dtype=int)
    chroms = np.array([], dtype="U25")
    for chrom in peaks:
        new_starts = np.arange(0, self.chrom_sizes[chrom]+bin_size, bin_size)
        starts = np.append(starts, new_starts)
        chroms = np.append(chroms, np.repeat(chrom, len(new_starts)))
    ends = starts + bin_size
    p_dist = IntervalFrame.from_array(starts, ends, labels=chroms)

    # Iterate over intervals
    p_dist.df.loc[:,"mean_dist"] = 0
    for i, interval in enumerate(p_dist.index):
        chrom = interval.label
        try:
            peaks[chrom]
            overlaps = peaks[chrom].intersect(interval.start, interval.end, label=chrom)
            if len(overlaps) > 3:
                p_dist.df.loc[i,"mean_dist"] = np.mean(overlaps.extract_ends()[:-1] - overlaps.extract_starts()[1:])
        except KeyError:
            pass

    # Remove those greater than max_distance
    p_dist = p_dist.iloc[p_dist.df.loc[:,"mean_dist"].values < max_distance, :]

    return p_dist


import ngsfragments as ngs


def wps(frags, protection=120, min_length=120, max_length=210, normalize=True, norm_method="mean"):
    """
    Calculate Window Protection Score (WPS)

    Parameters
    ----------
        chrom : str
            Name of chromosome
        protection : int
            Protection window
        min_length : int
            Minimum DNA fragment length
        max_length : int
            Maximum fragment length
        normalize : bool
            Whether to normalize
        norm_method : str
            Normalization method

    Returns
    -------
        wps : pandas.Series
            Window protection scores
    """
    
    # Caclulate WPS
    wps = frags.wps(protection=protection, min_length=min_length, max_length=max_length)
    if normalize:
        ngs.utilities.normalize_wps(wps, method=norm_method)

    return wps



import pandas as pd
import numpy as np
import h5py
from intervalframe.read.read_h5 import read_h5_intervalframe


def read_h5_DataFrame(h5_group):
    """
    Read pandas.DataFrame from h5py group

    Parameters
    ----------
        h5_group
            h5py.Group

    Returns
    -------
        df
            pandas.DataFrame
    """

    # Define dtype categories
    numeric_dtypes = set(['i','u','b','c'])
    string_dtypes = set(['O','U','S'])

    # Record axis
    axis = int(np.array(h5_group["axis"]))
    shape = np.array(h5_group["shape"])

    # Record columns type
    col_type = np.array(h5_group["columns_type"])[0].decode()

    # Read index
    if axis == 0:
        df = pd.DataFrame.from_records(np.array(h5_group["values"]), index="index")
        if df.index.dtype.kind in string_dtypes:
            df.index = df.index.values.astype(str)
        
    # Read columns
    elif axis == 1:
        if col_type == "single":
            df = pd.DataFrame.from_records(np.array(h5_group["values"]), index="index").T
            if df.columns.dtype.kind in string_dtypes:
                df.columns = df.columns.values.astype(str)
        elif col_type == "multi":
            df = pd.DataFrame.from_records(np.array(h5_group["values"]), index=["level_0","level_1"]).T
            df.columns = pd.MultiIndex.from_tuples([(a.decode(), b.decode()) for a,b in df.columns.values],
                                                    names=["level_0","level_1"])

    # Convert numpy object to str
    for i, dtype in enumerate(df.dtypes):
        if dtype.kind == "O":
            df.iloc[:,i] = df.iloc[:,i].values.astype(str)
        
    return df


def read_h5(cfdna_object, h5_filename):
    """
    """

    # Open file
    f = h5py.File(h5_filename, "r")

    # Read anno
    cfdna_object._anno = read_h5_DataFrame(f["anno"])

    # Read intervals
    for key in list(f["intervals"].keys()):
        cfdna_object._intervals[key] = read_h5_intervalframe(f["intervals"][key])

    # Read obs intervals
    for key in list(f["obs_intervals"]):
        cfdna_object._obs_intervals[key] = {}
        for obs in list(f["obs_intervals"][key]):
            cfdna_object._obs_intervals[key][obs] = read_h5_intervalframe(f["obs_intervals"][key][obs])

    # Read obs values
    for key in list(f["obs_values"]):
        cfdna_object._obs_values[key] = read_h5_DataFrame(f["obs_values"][key])

    # Read chroms
    cfdna_object.chroms = set(np.array(f["chroms"]).astype(str))

    # Read hmm
    hmm_states = {}
    for key in list(f["hmm_states"]):
        hmm_states[key] = {}
        hmm_states[key]["mus"] = np.array(f["hmm_states"][key]["mus"])
        hmm_states[key]["lambdas"] = np.array(f["hmm_states"][key]["lambdas"])
        hmm_states[key]["nu"] = np.array(f["hmm_states"][key]["nu"])[0]
        hmm_states[key]["states"] = pd.Series(np.array(f["hmm_states"][key]["states"]), name="Sample_1")
        hmm_states[key]["subclone"] = pd.Series(np.array(f["hmm_states"][key]["subclone"]), name="Sample_1")
        hmm_states[key]["n"] = np.array(f["hmm_states"][key]["n"])[0]
        hmm_states[key]["phi"] = np.array(f["hmm_states"][key]["phi"])[0]
        hmm_states[key]["sp"] = np.array(f["hmm_states"][key]["sp"])[0]
        hmm_states[key]["Frac_genome_subclonal"] = np.array(f["hmm_states"][key]["Frac_genome_subclonal"])[0]
        hmm_states[key]["cn"] = read_h5_intervalframe(f["hmm_states"][key]["cn"])
    cfdna_object.hmm_states = hmm_states

    # Read file names
    cfdna_object.filenames = set(np.array(f["filenames"]).astype(str))

    # Read bin sizes
    if "cnv_binsize" in list(f.keys()):
        cfdna_object.cnv_binsize = np.array(f["cnv_binsize"])[0]
    if "hmm_binsize" in list(f.keys()):
        cfdna_object.hmm_binsize = np.array(f["hmm_binsize"])[0]
    if "profile_binsize" in list(f.keys()):
        cfdna_object.profile_binsize = np.array(f["profile_binsize"])[0]

    # Read chrom lengths
    chroms = np.array(f["chrom_lengths"]["chroms"]).astype(str)
    lengths = np.array(f["chrom_lengths"]["lengths"])
    cfdna_object.chrom_lengths = {chrom:lengths[i] for i, chrom in enumerate(chroms)}

    # Close file
    f.close()


from __future__ import absolute_import
from .core.cfDNA import cfDNA
from . import plot as pl
from . import io
from . import commandline as cmd
from . import utilities as utils
from . import processing as proc

from ngsfragments import fragments


# This is extracted automatically by the top-level setup.py.
__version__ = '1.0.3'

__author__ = "Kyle S. Smith"

__doc__ = """\

API
======

Central class
-------------

.. autosummary::
   :toctree: .
   
   cfDNA
   
   
Utility functions
-----------------

.. autosummary::
   :toctree: .
   
   utils.h5_utilities


Plotting functions
------------------
    
.. autosummary::
   :toctree: .
   
   pl.plot_plt
   pl.plot_bokeh


IO functions
------------------
    
.. autosummary::
   :toctree: .
   
   io.write
   io.read

"""


from ngsfragments import fragments


def readBam(bam_filename, min_size=10, max_size=1000,
            paired=True, qcfail=False, mapq_cutoff=25,
            proportion=1.0, verbose=False):
    """
    Read BAM file

    Parameters
    ----------
        bam_filename : str
            Name of the BAM file to read
        min_size : int
            Minimum length of fragment to consider
        max_size : int
            Maximum length of fragment to consider
        paired : bool
            Whether to process as paired reads
        qcfail : bool
            Whether to keep QC failed reads
        mapq_cutoff : int
            Minimum mapping quality allowed
        proportion : float
            Proportion of reads to keep
        verbose : bool
            Print process
    
    Returns
    -------
        fragments : :class: `ngsfragments.ngsfragments`
            DNA fragment intervals from BAM file

    """

    # Read fragments from bam file
    frags = fragments(bam_filename, min_size=min_size, max_size=max_size,
                        paired=paired, qcfail=qcfail, mapq_cutoff=mapq_cutoff, verbose=verbose,
                        proportion=proportion)

    return frags


from ..core.cfDNA import cfDNA
from ..plot.plot_plt import summary, plot_cnv
from ..io.write.write_text import write_cnv_seg
from ..io.read.read_bam import readBam
from ..processing.cnv.segmentation import call_cnv_pipline
from ..processing.summarize.summarize import summarize
from ..utilities.h5_utilities import merge_h5
from ..io.write.write_h5 import write_h5
import os


def CNV_calling(args):
    """
    """

    # Detect if multiple bams
    if isinstance(args.bam, list):
        for bam in args.bam:
            if args.prefix == "":
                prefix = os.path.split(bam)[-1]
            else:
                prefix = args.prefix
            # Read bam
            frags = readBam(bam, min_size=args.min_length, max_size=args.max_length,
                                paired=args.single, qcfail=args.noqcfail, mapq_cutoff=args.mapq,
                                proportion=args.proportion, verbose=True)

            # Check downsampling
            if args.n_frags != 0:
                frags = frags.downsample(n_frags=args.n_frags)
            elif args.proportion != 1.0:
                frags = frags.downsample(args.proportion)

            # Create cfDNA object
            cfdna_object = cfDNA()
            #cfDNA_object = cfDNA(frags, ref_genome=args.genome, verbose=True)
            
            # Run CNV for hmm models
            call_cnv_pipline(cfdna_object, frags, cnv_binsize=args.bin_size, hmm_binsize=args.bin_size,
                                          method="bcp_online_both", outlier_smooth=True,
                                          gauss_smooth=False, bcp_cutoff=0.3, normal=[0.1, 0.5, 0.9],
                                          ploidy=[2], estimatePloidy=False, minSegmentBins=25,
                                          maxCN=7, verbose=False)
            
            # Write seg file
            if args.segs:
                print("Saving segmentation file:", prefix+".segs")
                seg_fn = prefix + ".segs"
                sample = os.path.split(prefix)[-1]
                write_cnv_seg(cfdna_object, sample, seg_fn)

            # Plot
            plot_cnv(cfdna_object, list(cfdna_object.filenames)[0], title=list(cfdna_object.filenames)[0],
                    show=False, save=prefix+"_cnv_plot.png", ax=None)

    else:
        if args.prefix == "":
            prefix = os.path.split(args.bam)[-1]
        else:
            prefix = args.prefix
        # Read bam
        frags = readBam(args.bam, min_size=args.min_length, max_size=args.max_length,
                            paired=args.single, qcfail=args.noqcfail, mapq_cutoff=args.mapq,
                            proportion=args.proportion, verbose=True)

        # Check downsampling
        if args.n_frags != 0:
            frags = frags.downsample(n_frags=args.n_frags)
        elif args.proportion != 1.0:
            frags = frags.downsample(args.proportion)
        
        # Create cfDNA object
        cfdna_object = cfDNA()
        #cfDNA_object = cfDNA(fragments, ref_genome=args.genome, verbose=True)

        # Run CNV for hmm models
        call_cnv_pipline(cfdna_object, frags, cnv_binsize=args.bin_size, hmm_binsize=args.bin_size,
                                        method="bcp_online_both", outlier_smooth=True,
                                        gauss_smooth=False, bcp_cutoff=0.3, normal=[0.1, 0.5, 0.9],
                                        ploidy=[2], estimatePloidy=False, minSegmentBins=25,
                                        maxCN=7, verbose=False)

        # Write seg file
        if args.segs:
            seg_fn = prefix + ".segs"
            sample = os.path.split(prefix)[-1]
            write_cnv_seg(cfdna_object, sample, seg_fn)

        # Plot
        plot_cnv(cfdna_object, list(cfdna_object.filenames)[0], title=list(cfdna_object.filenames)[0],
                 show=False, save=prefix+"_cnv_plot.png", ax=None)


def cfDNA_summarize(args):
    """
    """

    # Detect if multiple bams
    if isinstance(args.bam, list):
        for bam in args.bam:
            # Define prefix
            if args.prefix == "":
                prefix = os.path.split(bam)[-1]
            else:
                prefix = args.prefix + os.path.split(bam)[-1]

            # Define sample
            sample = os.path.split(prefix)[-1]

            # Create cfDNA object
            cfdna_object = cfDNA()
            frags = readBam(bam, proportion=args.proportion,
                                 verbose=True, min_size=args.min_length, max_size=args.max_length)

            # Check downsampling
            if args.n_frags != 0:
                frags = frags.downsample(n_frags=args.n_frags)
            elif args.proportion != 1.0:
                frags = frags.downsample(args.proportion)

            # Run summary
            summarize(cfdna_object, frags, args.bin_size, progress_bar=args.progressbar)

            # Plot metrics
            print("Saving summary plot:", prefix+"_summary.pdf")
            summary(cfdna_object, sample, show=False, save=prefix+"_summary.pdf")

            # Write metrics
            print("Saving summary h5:", prefix+"_summary.h5")
            write_h5(cfdna_object, prefix+"_summary.h5")
            
            # Write seg file
            if args.segs:
                print("Saving segmentation file:", prefix+".segs")
                seg_fn = prefix + ".segs"
                sample = os.path.split(prefix)[-1]
                write_cnv_seg(cfdna_object, sample, seg_fn)

    else:
        # Define BAM
        bam = args.bam

        # Define prefix
        if args.prefix == "":
            prefix = os.path.split(bam)[-1]
        else:
            prefix = args.prefix + os.path.split(bam)[-1]

        # Define sample
        sample = os.path.split(prefix)[-1]

        # Create cfDNA object
        cfdna_object = cfDNA()
        frags = readBam(bam, proportion=args.proportion,
                                verbose=True, min_size=args.min_length, max_size=args.max_length)

        # Check downsampling
        if args.n_frags != 0:
            frags = frags.downsample(n_frags=args.n_frags)
        elif args.proportion != 1.0:
            frags = frags.downsample(args.proportion)

        # Run summary
        summarize(cfdna_object, frags, args.bin_size, progress_bar=args.progressbar)

        # Plot metrics
        print("Saving summary plot:", prefix+"_summary.pdf")
        summary(cfdna_object, sample, show=False, save=prefix+"_summary.pdf")

        # Write metrics
        print("Saving summary h5:", prefix+"_summary.h5")
        write_h5(cfdna_object, prefix+"_summary.h5")
        
        # Write seg file
        if args.segs:
            print("Saving segmentation file:", prefix+".segs")
            seg_fn = prefix + ".segs"
            sample = os.path.split(prefix)[-1]
            write_cnv_seg(cfdna_object, sample, seg_fn)


def cfDNA_merge(args):
    """
    Merge multiple summary h5 file into one
    """

    # Detect directory
    directory_split = os.path.split(args.inputs[0])
    if directory_split[0] == "":
        directory = "."
    else:
        directory = directory_split[0]

    # Merge h5 files
    merge_h5(args.inputs, args.output, progress_bar=True)


import argparse
from .commandline.commands import cfDNA_summarize, cfDNA_merge, CNV_calling


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers()

    # Summarize
    parser_summary = subparsers.add_parser("summarize", help="Calculate, plot, and save summary metrics")
    parser_summary.add_argument("--bam", help="BAM file", required=True, nargs='*')
    parser_summary.add_argument("--prefix", help="Prefix for ouput files", default="")
    parser_summary.add_argument("--bin_size", type=int, help="Bin size to use (default=100000)", default=100000)
    parser_summary.add_argument("--genome", help="Version of genome to use (default=hg19)", default="hg19")
    parser_summary.add_argument("--proportion", type=float, help="Proportion of fragments to use (default: 1.0)", default=1.0)
    parser_summary.add_argument("--n_frags", type=int, help="Estimate of number of fragments to use (default: ALL)", default=0)
    parser_summary.add_argument("--min_length", type=int, help="Minimum length to consider (default: 1)", default=1)
    parser_summary.add_argument("--max_length", type=int, help="Maximum length to consider (default: 1000)", default=1000)
    parser_summary.add_argument("--segs", help="Whether to write seg files (default: False)", default=False, action="store_true")
    parser_summary.add_argument("--progressbar", help="Whether show progress bar (default: False)", default=False, action="store_true")
    parser_summary.set_defaults(func=cfDNA_summarize)

    # Call CNVs
    parser_summary = subparsers.add_parser("callCNVs", help="Plot and write CNV segments")
    parser_summary.add_argument("--bam", help="BAM file", required=True, nargs='*')
    parser_summary.add_argument("--prefix", help="Prefix for ouput files", default="")
    parser_summary.add_argument("--bin_size", type=int, help="Bin size to use (default=100000)", default=100000)
    parser_summary.add_argument("--hmm_bin_size", type=int, help="Bin size to use (default=1000000)", default=1000000)
    parser_summary.add_argument("--genome", help="Version of genome to use (default=hg19)", default="hg19")
    parser_summary.add_argument("--proportion", type=float, help="Proportion of fragments to use (default: 1.0)", default=1.0)
    parser_summary.add_argument("--n_frags", type=int, help="Estimate of number of fragments to use (default: ALL)", default=0)
    parser_summary.add_argument("--min_length", type=int, help="Minimum length to consider (default: 1)", default=1)
    parser_summary.add_argument("--max_length", type=int, help="Maximum length to consider (default: 1000)", default=1000)
    parser_summary.add_argument("--mapq", type=int, help="Mapping quality cutoff (default: 25)", default=25)
    parser_summary.add_argument("--segs", help="Whether to write seg files (default: False)", default=False, action="store_true")
    parser_summary.add_argument("--single", help="Whether to reads are single ended (default: False)", default=False, action="store_true")
    parser_summary.add_argument("--noqcfail", help="Whether to to remove qcfail flag (default: False)", default=False, action="store_true")
    parser_summary.set_defaults(func=CNV_calling)

    # Merge summaries
    parser_merge = subparsers.add_parser("merge", help="Calculate, plot, and save summary metrics")
    parser_merge.add_argument("--inputs", help="Directory for h5 summary files", required=True, nargs='*')
    parser_merge.add_argument("--output", help="h5 file name for output summary", required=True)
    parser_merge.set_defaults(func=cfDNA_merge)

    args = parser.parse_args()

    args.func(args)


import ngsfragments as ngs
import numpy as np
import pandas as pd
import ngsfragments as ngs
import os


def fragment_profile(cfdna_object, frags, bin_size=100000, bin_bias_h5_fn=None, bias_correct=True, smooth=True, verbose=False):
   """
   Determine fragment profile (ratio of large fragments to small fragments)

   Parameters
   ----------
      frags : ngsfragments.fragments
      chrom
            str
      bins_size
            int
      genome
            str
      bin_bias_h5_fn
            str
      bias_correct
            bool
      smooth
            bool

   Returns
   -------
      fragment_profile
            pandas.DataFrame
   """

   # Determine file name
   path = os.path.normpath(frags.sam_file)
   file_name = path.split(os.sep)[-1]

   # Check if file was previously annotated
   if file_name not in cfdna_object.filenames:
      cfdna_object.log_filename(frags)

   # Calculate coverage of small and large fragments
   small_bin_coverage = frags.bin_counts(bin_size=bin_size, min_length=100, max_length=150)
   large_bin_coverage = frags.bin_counts(bin_size=bin_size, min_length=151, max_length=220)

   # Correct coverage
   if bias_correct:
      small_bin_coverage = ngs.segment.correction.correct(small_bin_coverage, genome=cfdna_object.ref_genome, bin_size=bin_size, verbose=verbose)
      large_bin_coverage = ngs.segment.correction.correct(large_bin_coverage, genome=cfdna_object.ref_genome, bin_size=bin_size, verbose=verbose)
      small_bin_coverage = small_bin_coverage.loc[:,["corrected_counts"]]
      small_bin_coverage.df.columns = [file_name]
      large_bin_coverage = large_bin_coverage.loc[:,["corrected_counts"]]
      large_bin_coverage.df.columns = [file_name]
   else:
      small_bin_coverage.df.columns = [file_name]
      large_bin_coverage.df.columns = [file_name]

   # Calculate fragment profile
   fragment_profile = small_bin_coverage.copy()
   fragment_profile.df.loc[:,file_name] = small_bin_coverage.df.loc[:,file_name].values / large_bin_coverage.df.loc[:,file_name].values

   # Correct regions with 0 for both
   fragment_profile.df.loc[pd.isnull(fragment_profile.loc[:,file_name].values), file_name] = 1.0
   
   # Append bins
   cfdna_object.add_intervals("frag_profile", file_name, fragment_profile)


import numpy as np
import pandas as pd


def write_seg(iframe, filename, sample_name, value_key, binsize=100000):
    """
    """
    
    # Create pandas dataframe
    chroms = iframe.index.extract_labels()
    starts = iframe.starts()
    ends = iframe.ends()
    nbins = (ends - starts) / binsize
    seg_csv = pd.DataFrame([np.repeat(sample_name, iframe.shape[0]), chroms,
                            starts, ends, nbins.astype(int), iframe.df.loc[:,value_key]],
                           index=["sample","chrom","start","end","n_bins","value"]).T
    
    # Write
    seg_csv.to_csv(filename, header=True, index=None, sep="\t")


def write_cnv_seg(cfdna_object, key, filename):
    """
    """

    # Write seg file
    write_seg(cfdna_object.obs_intervals["cnv_segments"][key],
              filename, key, "median", cfdna_object.cnv_binsize)


from __future__ import division, print_function, absolute_import

# Python 3
MyFileNotFoundError = FileNotFoundError

# Library name
libname = "cfdna"

# Descriptions of package
SHORTDESC = "Python package for fragment manipulation for cfDNA"
#DESC = """A python package for fragment manipulation for cfDNA."""
with open("README.md", "r") as fh:
    long_description = fh.read()

# Directories (relative to the top-level directory where setup.py resides) in which to look for data files.
datadirs  = ("tests", "data")
# File extensions to be considered as data files. (Literal, no wildcards.)
dataexts  = (".py", ".pyx", ".pxd", ".c", ".h", ".h5", ".txt", ".bed")
# Standard documentation to detect (and package if it exists).
standard_docs     = ["README", "LICENSE", "TODO", "CHANGELOG", "AUTHORS"]
standard_doc_exts = [".md", ".rst", ".txt", ""]

#########################################################
# Init
#########################################################

# check for Python 2.7 or later
import sys
if sys.version_info < (2,7):
    sys.exit('Sorry, Python < 2.7 is not supported')

import os

from setuptools import setup, find_namespace_packages, find_packages


# Gather user-defined data files
datafiles = []
getext = lambda filename: os.path.splitext(filename)[1]
for datadir in datadirs:
    datafiles.extend( [(root, [os.path.join(root, f) for f in files if getext(f) in dataexts])
                       for root, dirs, files in os.walk(datadir)] )

# Add standard documentation (README et al.), if any, to data files
detected_docs = []
for docname in standard_docs:
    for ext in standard_doc_exts:
        filename = "".join( (docname, ext) )  # relative to the directory in which setup.py resides
        if os.path.isfile(filename):
            detected_docs.append(filename)
datafiles.append( ('.', detected_docs) )


# Extract __version__ from the package __init__.py
import ast
init_py_path = os.path.join(libname, '__init__.py')
version = '0.0.0'
try:
    with open(init_py_path) as f:
        for line in f:
            if line.startswith('__version__'):
                version = ast.parse(line).body[0].value.s
                break
        else:
            print( "WARNING: Version information not found in '%s', using placeholder '%s'" % (init_py_path, version), file=sys.stderr )
except MyFileNotFoundError:
    print( "WARNING: Could not find file '%s', using placeholder version information '%s'" % (init_py_path, version), file=sys.stderr )


#########################################################
# Call setup()
#########################################################

setup(
    name = "cfdna",
    version = version,
    author = "Kyle S. Smith",
    author_email = "kyle.smith@stjude.org",
    url = "https://github.com/kylessmtih/cfdna",
    description = SHORTDESC,
    long_description = long_description,
    long_description_content_type = "text/markdown",
    # CHANGE THIS
    license = "GPL2",
    # free-form text field
    platforms = ["Linux"],
    classifiers = [ "Development Status :: 4 - Beta",
                    "Environment :: Console",
                    "Intended Audience :: Developers",
                    "Intended Audience :: Science/Research",
                    "Operating System :: POSIX :: Linux",
                    "Programming Language :: Python",
                    "Programming Language :: Python :: 3",
                    "Programming Language :: Python :: 3.4",
                    "Topic :: Scientific/Engineering",
                    "Topic :: Scientific/Engineering :: Mathematics",
                    "Topic :: Software Development :: Libraries",
                    "Topic :: Software Development :: Libraries :: Python Modules",
                    "Topic :: Scientific/Engineering :: Bio-Informatics"
                  ],
    setup_requires = ["numpy","ngsfragments","intervalframe","matplotlib","seaborn","bokeh","hmmCNV"],
    install_requires = ["numpy","ngsfragments","h5py","pandas"],
    provides = ["cfdna"],
    keywords = ["next generation sequencing fragment cfDNA cell free DNA"],
    packages = find_packages(),
    package_data={'cfdna': ['*.pxd', '*.pyx', '*.c', '*.h']},
    # Disable zip_safe
    zip_safe = False,
    # Custom data files not inside a Python package
    data_files = datafiles
)


import ngsfragments as ngs
import os
import numpy as np
np.seterr(all="ignore")
#from ngsfragments.segment.cnv import call_cnvs, process_cnvs
#from ngsfragments.segment import cnv_utilities


def call_cnvs(cfdna_object, frags, bin_size=100000, bin_bias_h5_fn=None, method="bcp_online_both",
                  outlier_smooth=True, gauss_smooth=False, bcp_cutoff=0.3, verbose=False, **kwargs):
    """
    Call Copy Number Variants (CNVs)

    Parameters
    ----------
    chrom : str
        Name of chromosome
    bin_size : int
    min_length : int
    max_length : int
    genome : str
    bin_bias_h5_fn : str
    genome_fn : str
    centro_fn : str
    detail_fn : str
    outlier_smooth : bool
    gauss_smooth : bool
    normal : list
    ploidy : list
    estimatePloidy : bool
    minSegmentBins : int
    maxCN : int
    bcp_cutoff : float
        
    Returns
    -------
        cnv : :class:`pandas.DataFrame`

    """

    # Find file name
    path = os.path.normpath(frags.sam_file)
    file_name = path.split(os.sep)[-1]

    # Check if file was previously annotated
    if file_name not in cfdna_object.filenames:
        cfdna_object.log_filename(frags)

    # Determine bin size
    if cfdna_object.cnv_binsize is None:
        cfdna_object.cnv_binsize = bin_size
    elif cfdna_object.cnv_binsize != bin_size:
        raise AttributeError("Provided bin_size does not match previous bin_size")

    # Call Copy Number Variations
    cnv_bins, cnv_segs = ngs.segment.cnv.call_cnvs(frags, bin_size=bin_size, bin_bias_h5_fn=bin_bias_h5_fn,
                                                    genome=cfdna_object.ref_genome, method=method, outlier_smooth=outlier_smooth,
                                                    gauss_smooth=gauss_smooth, verbose=verbose, cutoff=bcp_cutoff)
    
    # Append bins
    cnv_bins = cnv_bins.loc[:,["ratios"]]
    cnv_bins.df.columns = [file_name]
    cfdna_object.add_obs_intervals("cnv_segments", file_name, cnv_segs)
    cfdna_object.add_intervals("cnv_bins", file_name, cnv_bins)

    #if cfdna_object.cnv_bins is None:
    #    cfdna_object.cnv_bins = cnv_bins
    #else:
    #    cnv_bins = cnv_bins.exact_match(cfdna_object.cnv_bins)
    #    cfdna_object.cnv_bins = cfdna_object.cnv_bins.exact_match(cnv_bins)
    #    cfdna_object.cnv_bins.df.loc[:,file_name] = cnv_bins.loc[:,file_name].values


def predict_purity(cfdna_object, frags, bin_size = 1000000, bin_bias_h5_fn=None, method="bcp_online_both",
                    outlier_smooth=True, gauss_smooth=False, bcp_cutoff=0.3, normal = [0.1, 0.5, 0.9],
                    ploidy = [1, 2, 3], estimatePloidy=False, minSegmentBins=25, maxCN=7, verbose=False, **kwargs):
    """
    Predict tumor purity
    """

    # Find file name
    path = os.path.normpath(frags.sam_file)
    file_name = path.split(os.sep)[-1]

    # Check if file was previously annotated
    if file_name not in cfdna_object.filenames:
        cfdna_object.log_filename(frags)

    # Calculate bins
    hmm_binsize = bin_size
    bins, segments = ngs.segment.cnv.call_cnvs(frags, bin_size=bin_size, bin_bias_h5_fn=bin_bias_h5_fn, genome=cfdna_object.ref_genome, method=method,
                                                outlier_smooth=outlier_smooth, gauss_smooth=gauss_smooth, verbose=verbose, cutoff=bcp_cutoff)

    # Define gender
    try:
        gender = cfdna_object.gender[file_name]
    except KeyError:
        gender = None
    
    # Classify calls be HMM
    hmm_states = ngs.segment.cnv_utilities.train_hmm(bins, normal=normal, ploidy=ploidy,
                                                    gender=gender, estimatePloidy=estimatePloidy,
                                                    minSegmentBins=minSegmentBins, maxCN=maxCN,
                                                    verbose=verbose, **kwargs)

    # Assign attributes
    cfdna_object.add_anno("purity", file_name, 1 - hmm_states["n"])
    cfdna_object.add_anno("ploidy", file_name, hmm_states["phi"])
    cfdna_object.add_anno("clonal", file_name, hmm_states["Frac_genome_subclonal"])
    #cfdna_object.purity[file_name] = 1 - hmm_states["n"]
    #cfdna_object.ploidy[file_name] = hmm_states["phi"]
    #cfdna_object.clonal[file_name] = hmm_states["Frac_genome_subclonal"]


def train_hmm(cfdna_object, frags, bin_size = 1000000, bin_bias_h5_fn=None, method="bcp_online_both",
                outlier_smooth=True, gauss_smooth=False, bcp_cutoff=0.3, normal = [0.1, 0.5, 0.9],
                ploidy = [2], estimatePloidy=False, minSegmentBins=25, maxCN=7, verbose=False, **kwargs):
    """
    Train Hidden Markov Model (HMM)
    """

    # Determine file name
    path = os.path.normpath(frags.sam_file)
    file_name = path.split(os.sep)[-1]

    # Check if file was previously annotated
    if file_name not in cfdna_object.filenames:
        cfdna_object.log_filename(frags)

    # Determine bin size
    if cfdna_object.hmm_binsize is None:
        cfdna_object.hmm_binsize = bin_size
    elif cfdna_object.hmm_binsize != bin_size:
        raise AttributeError("Provided bin_size does not match previous bin_size")

    # Calculate bins
    bins, segments = ngs.segment.cnv.call_cnvs(frags, bin_size=bin_size, bin_bias_h5_fn=bin_bias_h5_fn, genome=cfdna_object.ref_genome, method=method,
                                outlier_smooth=outlier_smooth, gauss_smooth=gauss_smooth, verbose=verbose, cutoff=bcp_cutoff)

    # Define gender
    try:
        gender = cfdna_object.gender[file_name]
    except KeyError:
        gender = None

    # Train HMM
    cfdna_object.hmm_states[file_name] = ngs.segment.cnv_utilities.train_hmm(bins, normal=normal, ploidy=ploidy,
                                                                gender=gender, estimatePloidy=estimatePloidy,
                                                                minSegmentBins=minSegmentBins, maxCN=maxCN,
                                                                verbose=verbose, **kwargs)



def process_cnvs(cfdna_object, frags=None, file_name=None, merge=True):
    """
    """

    # Find file_name
    if frags is None and file_name is None:
        raise NameError("frags or file_name must be provided.")
    elif frags is not None:
        path = os.path.normpath(frags.sam_file)
        file_name = path.split(os.sep)[-1]

        # Check if file was previously annotated
        if file_name not in cfdna_object.filenames:
            cfdna_object.log_filename(frags)
    
    # Check if hmm has been run
    try:
        cfdna_object.hmm_states[file_name]
    except KeyError:
        raise AttributeError("Must run .train_hmm() before .process_cnvs()")

    # Extract sample bins
    cfdna_sample = cfdna_object.intervals["cnv_bins"].loc[:,[file_name]]
    cfdna_sample.df.columns = ["ratios"]

    # Merge segments
    if merge:
        cfdna_object.obs_intervals["cnv_segments"][file_name] = ngs.segment.cnv_utilities.merge_segments(cfdna_object.obs_intervals["cnv_segments"][file_name],
                                                                                        cfdna_sample)

    # Process CNVs
    cfdna_object.obs_intervals["cnv_segments"][file_name] = ngs.segment.cnv_utilities.hmm_classify(cfdna_object.obs_intervals["cnv_segments"][file_name],
                                                                                  cfdna_object.hmm_states[file_name])
    ngs.segment.cnv_utilities.validate_distributions(cfdna_object.obs_intervals["cnv_segments"][file_name],
                                                     cfdna_sample)


def call_cnv_pipline(cfdna_object, frags, cnv_binsize=100000, hmm_binsize=1000000, method="bcp_online_both", outlier_smooth=True,
                         gauss_smooth=False, bcp_cutoff=0.3, normal=[0.1, 0.5, 0.9], ploidy=[2], estimatePloidy=False, minSegmentBins=25,
                         maxCN=7, verbose=False):
    """
    """

    # Determine file name
    path = os.path.normpath(frags.sam_file)
    file_name = path.split(os.sep)[-1]

    # Check if file was previously annotated
    if file_name not in cfdna_object.filenames:
        cfdna_object.log_filename(frags)

    # Predict purity and ploidy
    predict_purity(cfdna_object, frags, bin_size = hmm_binsize, bin_bias_h5_fn=None, method=method,
                        outlier_smooth=outlier_smooth, gauss_smooth=gauss_smooth, bcp_cutoff=bcp_cutoff, normal=normal,
                        ploidy=ploidy, estimatePloidy=estimatePloidy, minSegmentBins=minSegmentBins, maxCN=maxCN, verbose=verbose)

    # Train HMM
    train_hmm(cfdna_object, frags, bin_size = hmm_binsize, bin_bias_h5_fn=None, method=method,
                outlier_smooth=outlier_smooth, gauss_smooth=gauss_smooth, bcp_cutoff=bcp_cutoff, normal=normal,
                ploidy=ploidy, estimatePloidy=estimatePloidy, minSegmentBins=minSegmentBins, maxCN=maxCN, verbose=verbose)

    # Call Copy Number Variations
    call_cnvs(cfdna_object, frags, bin_size=cnv_binsize, bin_bias_h5_fn=None, method=method,
                outlier_smooth=outlier_smooth, gauss_smooth=gauss_smooth, verbose=verbose, cutoff=bcp_cutoff)
    process_cnvs(cfdna_object, frags)

    # Calculate segment variance
    cfdna_object.obs_intervals["cnv_segments"][file_name].annotate(cfdna_object.intervals["cnv_bins"], file_name, "var")